/* game.c    moving men around the board, enforcing the rules  */

/*
 * Dame:     a point and click checkerboard
 * (C):      1998  Peter Chiocchetti  <girbal@super.tacheles.de>
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#include <dame.h>


/*
 * Game Play
 */

/* Get the directions right.  */

static int
north (int y)
{
  switch (players.current) {
  case Black:
    if (direction == North)
      y -= 1;
    else
      y += 1;
    break;
  case White:
    if (direction == North)
      y += 1;
    else
      y -= 1;
    break;
  default:
    g_assert_not_reached();
  }
  return y;
}

static int
south (int y)
{
  switch (players.current) {
  case Black:
    if (direction == North)
      y += 1;
    else
      y -= 1;
    break;
  case White:
    if (direction == North)
      y -= 1;
    else
      y += 1;
    break;
  default:
    g_assert_not_reached();
  }
  return y;
}

static int
west (int x)
{
  switch (players.current) {
  case Black:
    if (direction == North)
      x -= 1;
    else
      x += 1;
    break;
  case White:
    if (direction == North)
      x += 1;
    else
      x -= 1;
    break;
  default:
    g_assert_not_reached();
  }
  return x;
}

static int
east (int x)
{
  switch (players.current) {
  case Black:
    if (direction == North)
      x += 1;
    else
      x -= 1;
    break;
  case White:
    if (direction == North)
      x -= 1;
    else
      x += 1;
    break;
  default:
    g_assert_not_reached();
  }
  return x;
}


/* This is about staying on board, hopping only over the enemie's stones,
   and settling only unoccupied ground.  */

static int
is_there (int x, int y)
{
  if (((x >= 0) && (x < CELLS)) && ((y >= 0) && (y < CELLS)))
    return (1);
  return (0);
}

static int
is_free (int x, int y)
{
  if (is_there(x, y)) {
    if (board[x][y].colour == 1)
      return (1);
  }
  return (0);
}

static int
is_opponent (int x, int y)
{
  if (is_there(x, y)) {
    switch (players.current) {
    case Black:
      if ((board[x][y].colour == 3) || (board[x][y].colour == 5))
	return (1);
      break;
    case White:
      if ((board[x][y].colour == 2) || (board[x][y].colour == 4))
	return (1);
      break;
    default:
      g_assert_not_reached();
    }
  }
  return (0);
}

static int
can_move (int x, int y)
{
  switch (board[x][y].colour) {
  case 5:
  case 4:
    if ((is_free(west(x), south(y))) || (is_free(east(x), south(y))))
      return (1);
  case 3:
  case 2:
    if ((is_free(west(x), north(y))) || (is_free(east(x), north(y))))
      return (1);
    break;
  default:
    g_assert_not_reached();
  }
  return (0);
}

static int
can_jump (int x, int y)
{
  switch (board[x][y].colour) {
  case 5:
  case 4:
    if (((is_opponent(west(x), south(y)))
	 && (is_free(west(west(x)), south(south(y)))))
	||
	((is_opponent(east(x), south(y)))
	 && (is_free(east(east(x)), south(south(y))))))
      return (1);
  case 3:
  case 2:
    if (((is_opponent(west(x), north(y)))
	 && (is_free(west(west(x)), north(north(y)))))
	||
	((is_opponent(east(x), north(y)))
	 && (is_free(east(east(x)), north(north(y))))))
      return (1);
    break;
  default:
    g_assert_not_reached();
  }
  return (0);
}


/* Scanning the whole board for jumps and moves,
   regarding whose turn it is.  */

static int
any_moves () /* Did the game start yet ?  */
{
  int x, y;

  for (y = 0; y < CELLS; y++) {
    for (x = 0; x < CELLS; x++) {
      switch (players.current) {
      case Black:
	if ((board[x][y].colour == 2) || (board[x][y].colour == 4)) {
	  if (can_move(x, y))
	    return (1);
	}
	break;
      case White:
	if ((board[x][y].colour == 3) || (board[x][y].colour == 5)) {
	  if (can_move(x, y))
	    return (1);
	}
	break;
      default:
	g_assert_not_reached();
      }
    }
  }
  return (0);
}

static int
any_jumps ()
{
  int x, y;

  for (y = 0; y < CELLS; y++) {
    for (x = 0; x < CELLS; x++) {
      switch (players.current) {
      case Black:
	if ((board[x][y].colour == 2) || (board[x][y].colour == 4)) {
	  if (can_jump(x, y))
	    return (1);
	}
	break;
      case White:
	if ((board[x][y].colour == 3) || (board[x][y].colour == 5)) {
	  if (can_jump(x, y))
	    return (1);
	}
	break;
      default:
	g_assert_not_reached();
      }
    }
  }
  return (0);
}
	    

/* Mark piece as touched by a human player; stick with it, only if
   it shall be moved - i.e. impose jumps.  */

static void
touch_stone_def (int x, int y)
{
  touched[0] = 1;
  touched[1] = board[x][y].colour;
  touched[2] = x;
  touched[3] = y;
}
    
static void
touch_stone (int x, int y)
{
  switch (players.current) {
  case Black:
    if ((board[x][y].colour == 2) || (board[x][y].colour == 4)) {
      if (must_jump) {
	if (can_jump(x, y))
	  touch_stone_def (x, y);
      }
      else {
	if (can_move(x, y))
	  touch_stone_def (x, y);
      }
    }
    break;
  case White:
    if ((board[x][y].colour == 3) || (board[x][y].colour == 5)) {
      if (must_jump) {
	if (can_jump(x, y))
	  touch_stone_def (x, y);
      }
      else {
	if (can_move(x, y))
	  touch_stone_def (x, y);
      }
    }
    break;
  default:
    g_assert_not_reached();
  }
}


/* Move pieces around the board, providing visual feedback.
   Take stones hopped over.  */

static void
move_stone (int x, int y, int xx, int yy)
{
  board[x][y].colour = 1;
  draw_cell (x, y);
  board[xx][yy].colour = saved_colour;
  draw_cell (xx, yy);
  if (must_jump) {
    board[(x + xx) / 2][(y + yy) / 2].colour = 1;
    draw_cell ((x + xx) / 2, (y + yy) / 2);
  }
}

/* Translate nodes to board x and y coordinates.  */

static void
locate_stone (int node, int *loc_x, int *loc_y)
{
  int x, y;

  for (y = 0; y < CELLS; y++) {
    for (x = 0; x < CELLS; x++) {
      if (board[x][y].number == node) {
	*loc_x = x;
	*loc_y = y;
	return;
      }
    }
  }
}

/* Connect start and end point of a single game turn given in standard
   notation.  The way from there to there in fact is resolved like a
   maze(1).  Only the piece to jump gets temporarily removed, not to
   be in its own way.  Will that be enough ?  The most horrible
   constellation seems to be jumping on the spot - i. e. in a circle.
   Also take care, not to jump right back and forth the same two
   fields. Touch pieces that can jump even further ! Crown kings, if
   a piece just reached the opposite last row, after all jumps have
   taken place.  Inform checkers engine about what happened.  Switch
   players.  */

void
trace_move (int home, int target)
{
  int count     = 0;
  int stopover  = 0;
  int i, j, power;

  int step_x[4] = { east(0) , west(0) , east(0) , west (0) };
  int step_y[4] = { north(0), north(0), south(0), south(0) };

  static struct {
    int x;
    int y;
    int dir;
  } route[MAXDEPTH];

  static int loc_x, loc_y;
  static int saved_loc;
  static char move[MAXDEPTH * 5 + 1];
  static ssize_t len;

  /* Memorize starting point.  */
  locate_stone (home, &loc_x, &loc_y);
  saved_colour     = board[loc_x][loc_y].colour;
  route[count].dir = power = i = saved_colour > 3 ? 4 : 2;
  route[count].x   = loc_x;
  route[count].y   = loc_y;
  if (!go_on) {
    saved_loc = board[loc_x][loc_y].number;
    len = sprintf (move, "%d", saved_loc);
  }

  /* Search the the board for this very move.  */
  /* Consider only jumps.  */
  if (must_jump) {
    board[loc_x][loc_y].colour = 1;
    while (stopover != target) {
      i--;
      if (i < 0) {
	count--;
	i = route[count].dir - 1;
	if (count < 0) {
	  /* No cheating, it's not a legal jump.  */
	  board[loc_x][loc_y].colour = saved_colour;
	  return;
	}
      }
      if (count >= MAXDEPTH - 1) {
	fprintf (stderr, _("Dame : too many consecutive jumps !\n"));
	board[loc_x][loc_y].colour = saved_colour;
	return;
      }
      route[count].dir = i;
      if ((is_opponent(route[count].x + step_x[i],
		       route[count].y + step_y[i]))
	  && (is_free(route[count].x + step_x[i] + step_x[i],
		      route[count].y + step_y[i] + step_y[i]))) {
	route[count + 1].x = route[count].x + step_x[i] + step_x[i];
	route[count + 1].y = route[count].y + step_y[i] + step_y[i];
	if ((route[count + 1].x == route[count - 1].x)
	    && (route[count + 1].y == route[count - 1].y))
	  continue;
	stopover = board[route[count + 1].x][route[count + 1].y].number;
	count++;
	i = power;
      }
    }
    board[loc_x][loc_y].colour = saved_colour;
    count--;
  }
  /* Consider only simple moves.  */
  else {
    while (stopover != target) {
      i--;
      if (i < 0)
	/* No cheating, it's an illegal move.  */
	return;
      if (is_free(route[count].x + step_x[i],
		  route[count].y + step_y[i])) {
	route[count + 1].x = route[count].x + step_x[i];
	route[count + 1].y = route[count].y + step_y[i];
	stopover = board[route[count + 1].x][route[count + 1].y].number;
      }
    }
  }

  /* Actually move stones.  Write down a path.  */
  for (j=0; j <= count; j++) {
    move_stone (route[j].x, route[j].y, route[j + 1].x, route[j + 1].y);
    len += sprintf (move + len, "-%d",
		    board[route[j + 1].x][route[j + 1].y].number);
  }

  /* Test whether all jumps have been used up.  */
  /* Force completion.  */
  if (must_jump && can_jump (route[count + 1].x, route[count + 1].y)) {
      go_on = 1;
      touch_stone_def (route[count + 1].x, route[count + 1].y);
    }
  /* Hand the board over.  */
  else {
    go_on = 0;
    touched[0] = 0;

    /* Switch players.  */
    /* Make kings.  */
    /* Tell move to the opponing engine, if there is any.  */
    switch (players.current) {
    case Black:
      players.current = White;
/*       gtk_window_set_title (GTK_WINDOW (app), _("Dame : White")); */
      if ((saved_colour == 2)
	  && (target > (CELLS * CELLS / 2 - CELLS / 2))) {
	board[route[count + 1].x][route[count + 1].y].colour = 4;
	draw_cell (route[count + 1].x, route[count + 1].y);
      }
      if (strcmp (players.white, username))
	tell_move (move);
      if (verbose)
	fprintf (stdout, "   %2d-%2d", saved_loc,
		 board[route[count + 1].x][route[count + 1].y].number);
      break;
    case White:
      players.current = Black;
/*       gtk_window_set_title (GTK_WINDOW (app), _("Dame : Black")); */
      if ((saved_colour == 3) && (target <= (CELLS / 2))) {
	board[route[count + 1].x][route[count + 1].y].colour = 5;
	draw_cell (route[count + 1].x, route[count + 1].y);
      }
      if (strcmp (players.black, username))
	tell_move (move);
      if (verbose)
	fprintf (stdout, "   %2d-%2d\n", saved_loc,
		 board[route[count + 1].x][route[count + 1].y].number);
      break;
    default:
      g_assert_not_reached();
    }

    /* Prepare for next turn.  */
    must_jump = any_jumps ();
    may_move  = any_moves ();

    /* Set statusbar  */
    switch (players.current) {
    case Black:
      if (must_jump)
	gnome_appbar_set_status (GNOME_APPBAR (ab),
				 _("Black's move - you must jump !"));
      else
	gnome_appbar_set_status (GNOME_APPBAR (ab), _("Black's move"));
      break;
    case White:
      if (must_jump)
	gnome_appbar_set_status (GNOME_APPBAR (ab),
				 _("White's move - you must jump !"));
      else
	gnome_appbar_set_status (GNOME_APPBAR (ab), _("White's move"));
      break;
    default:
      g_assert_not_reached();
    }

    /* Check for end of game.  */
    if ((!must_jump) && (!may_move)) {
      game_over = 1;
      switch (players.current) {
      case White:
	gnome_appbar_set_status (GNOME_APPBAR (ab), _("Black Wins"));
	if (verbose) {
	  fprintf (stdout, "\n");
	  fprintf (stdout, _("Black wins\n"));
	}
	break;
      case Black:
	gnome_appbar_set_status (GNOME_APPBAR (ab), _("White Wins"));
	if (verbose)
	  fprintf (stdout, _("White wins\n"));
	break;
      default:
	g_assert_not_reached();
      }
      stop_talk ();
    }
  }
}


/* Translate mouse clicks into game events.  */

void
turn_trap (int x, int y)
{
  switch (players.current) {
  case Black:
    if (strcmp (players.black, username))
      return;
    break;
  case White:
    if (strcmp (players.white, username))
      return;
    break;
  default:
    g_assert_not_reached();
  }
  if (!touched[0])
    touch_stone (x, y);
  else {
    trace_move (board[touched[2]][touched[3]].number, board[x][y].number);
  }
}
